from modules.agents import REGISTRY as agent_REGISTRY
from components.action_selectors import REGISTRY as action_REGISTRY
import torch as th
import torch.nn.functional as F
import torch.nn as nn
import numpy as np
# This multi-agent controller shares parameters between agents
class BasicMACExplore:
    def __init__(self, scheme, groups, args):
        self.n_agents = args.n_agents
        self.n_actions = args.n_actions
        self.state_dim = int(np.prod(args.state_shape))
        self.obs_shape = scheme["obs"]["vshape"]
        self.args = args
        self.rnn_hidden_dim = self.args.rnn_hidden_dim
        hypernet_embed = self.args.hypernet_embed
        input_shape = self._get_input_shape(scheme)
        self._build_agents(input_shape)
        self.agent_output_type = args.agent_output_type

        self.action_selector = action_REGISTRY[args.action_selector](args)
        self.hidden_states = None
        # self.curiosity_C = th.zeros([self.n_agents, self.rnn_hidden_dim, self.rnn_hidden_dim], dtype=th.float32)
        # self.influence_C = th.zeros([self.n_agents, self.rnn_hidden_dim, self.rnn_hidden_dim], dtype=th.float32)
        # self.explore_decay_alpha = self.args.explore_decay_alpha

        self.hyper_w_final = nn.Sequential(nn.Linear(self.state_dim, hypernet_embed),
                                           nn.ReLU(),
                                           nn.Linear(hypernet_embed, self.n_agents),
                                           nn.Sigmoid())


    def select_actions(self, ep_batch, t_ep, t_env, bs=slice(None), test_mode=False):
        # Only select actions for the selected batch elements in bs
        batch_size = ep_batch["avail_actions"].shape[0]
        avail_actions = ep_batch["avail_actions"][:, t_ep]
        agent_outputs, agent_hid, _, _, _ = self.forward_hid(ep_batch, t_ep, test_mode=test_mode) # [bs, n_agents, n_actions],[bs, n_agents, hid]
        

        # if test_mode == False:
        #     tmp_curiosity_C = self.curiosity_C.unsqueeze(0).repeat(batch_size,1,1,1).view(-1,self.rnn_hidden_dim,self.rnn_hidden_dim)  # bs*n_agents, rnn_hidden_dim, rnn_hidden_dim
        #     curiosity_reward = th.bmm(tmp_curiosity_C, tmp_hid.view(-1,self.rnn_hidden_dim,1)) # bs*agents, n_hid,1
        #     curiosity_reward = th.bmm(tmp_hid.view(-1,1,self.rnn_hidden_dim),curiosity_reward).view(batch,self.n_agents,1).repeat(1, 1, self.n_actions) # bs，agents,n_actions
        #     self.curiosity_C = (1 - self.explore_decay_alpha) * self.C + self.explore_decay_alpha

     

        chosen_actions = self.action_selector.select_action(agent_outputs[bs], avail_actions[bs], t_env, test_mode=test_mode)
        return chosen_actions

    def forward(self, ep_batch, t, test_mode=False):
        agent_inputs = self._build_inputs(ep_batch, t)
        avail_actions = ep_batch["avail_actions"][:, t]
        agent_outs, self.hidden_states = self.agent(agent_inputs, self.hidden_states)

        if self.agent_output_type == "pi_logits":

            if getattr(self.args, "mask_before_softmax", True):
                # Make the logits for unavailable actions very negative to minimise their affect on the softmax
                reshaped_avail_actions = avail_actions.reshape(ep_batch.batch_size * self.n_agents, -1)
                agent_outs[reshaped_avail_actions == 0] = -1e10

            agent_outs = th.nn.functional.softmax(agent_outs, dim=-1)
            if not test_mode:
                # Epsilon floor
                epsilon_action_num = agent_outs.size(-1)
                if getattr(self.args, "mask_before_softmax", True):
                    # With probability epsilon, we will pick an available action uniformly
                    epsilon_action_num = reshaped_avail_actions.sum(dim=1, keepdim=True).float()

                agent_outs = ((1 - self.action_selector.epsilon) * agent_outs
                               + th.ones_like(agent_outs) * self.action_selector.epsilon/epsilon_action_num)

                if getattr(self.args, "mask_before_softmax", True):
                    # Zero out the unavailable actions
                    agent_outs[reshaped_avail_actions == 0] = 0.0

        return agent_outs.view(ep_batch.batch_size, self.n_agents, -1)

    def forward_hid(self, ep_batch, t, test_mode=False):
        agent_inputs = self._build_inputs(ep_batch, t)
        avail_actions = ep_batch["avail_actions"][:, t]
        agent_outs, self.hidden_states = self.agent(agent_inputs, self.hidden_states)
     
        if self.agent_output_type == "pi_logits":

            if getattr(self.args, "mask_before_softmax", True):
                # Make the logits for unavailable actions very negative to minimise their affect on the softmax
                reshaped_avail_actions = avail_actions.reshape(ep_batch.batch_size * self.n_agents, -1)
                agent_outs[reshaped_avail_actions == 0] = -1e10

            agent_outs = th.nn.functional.softmax(agent_outs, dim=-1)
            if not test_mode:
                # Epsilon floor
                epsilon_action_num = agent_outs.size(-1)
                if getattr(self.args, "mask_before_softmax", True):
                    # With probability epsilon, we will pick an available action uniformly
                    epsilon_action_num = reshaped_avail_actions.sum(dim=1, keepdim=True).float()

                agent_outs = ((1 - self.action_selector.epsilon) * agent_outs
                               + th.ones_like(agent_outs) * self.action_selector.epsilon/epsilon_action_num)

                if getattr(self.args, "mask_before_softmax", True):
                    # Zero out the unavailable actions
                    agent_outs[reshaped_avail_actions == 0] = 0.0

        agent_outs = agent_outs.view(ep_batch.batch_size, self.n_agents, -1) 
        hidden_states = self.hidden_states.view(ep_batch.batch_size, self.n_agents, -1)
     
        states = avail_actions = ep_batch["state"][:, t]  # bs, state_dim
        w_final = self.hyper_w_final(states)              # bs, n_agent

        # w_final = th.zeros_like(w_final)
        # w_final[:,:] = 0.35
        # w_final[:,5] = 1 

        w_f = th.abs(w_final).unsqueeze(-1).repeat(1,1,self.n_actions)  # bs, n_agent, n_actions
        coop_agent_outs= w_f * agent_outs
        ind_agent_outs = (1 - w_f) * agent_outs
    
        # return agent_outs, hidden_states, ind_agent_outs, coop_agent_outs, w_final
        return agent_outs, hidden_states, agent_outs, agent_outs, w_final
    

    def init_hidden(self, batch_size):
        self.hidden_states = self.agent.init_hidden().unsqueeze(0).expand(batch_size, self.n_agents, -1)  # bav

    def agent_parameters(self):
        return self.agent.parameters()

    def w_parameters(self):
        return self.hyper_w_final.parameters()

    def parameters(self):
        return list(self.agent.parameters()) + list(self.hyper_w_final.parameters())

    def load_state(self, other_mac):
        self.agent.load_state_dict(other_mac.agent.state_dict())
        self.hyper_w_final.load_state_dict(other_mac.hyper_w_final.state_dict())
   


    def cuda(self):
        self.agent.cuda()
        self.hyper_w_final.cuda()


    def save_models(self, path):
        th.save(self.agent.state_dict(), "{}/agent.th".format(path))

    def load_models(self, path):
        self.agent.load_state_dict(th.load("{}/agent.th".format(path), map_location=lambda storage, loc: storage))

    def _build_agents(self, input_shape):
        self.agent = agent_REGISTRY[self.args.agent](input_shape, self.args)

    def _build_inputs(self, batch, t):
        # Assumes homogenous agents with flat observations.
        # Other MACs might want to e.g. delegate building inputs to each agent
        bs = batch.batch_size
        inputs = []
        inputs.append(batch["obs"][:, t])  # b1av
        if self.args.obs_last_action:
            if t == 0:
                inputs.append(th.zeros_like(batch["actions_onehot"][:, t]))
            else:
                inputs.append(batch["actions_onehot"][:, t-1])
        if self.args.obs_agent_id:
            inputs.append(th.eye(self.n_agents, device=batch.device).unsqueeze(0).expand(bs, -1, -1))

        inputs = th.cat([x.reshape(bs*self.n_agents, -1) for x in inputs], dim=1)
        return inputs

    def _get_input_shape(self, scheme):
        input_shape = scheme["obs"]["vshape"]
        if self.args.obs_last_action:
            input_shape += scheme["actions_onehot"]["vshape"][0]
        if self.args.obs_agent_id:
            input_shape += self.n_agents

        return input_shape
